package cn.minbb.serialport;

import android.os.Handler;
import android.os.HandlerThread;
import android.os.Message;
import android.util.Log;

import java.io.File;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Timer;
import java.util.TimerTask;

import cn.minbb.serialport.app.Data;
import cn.minbb.serialport.app.DataType;
import cn.minbb.serialport.handler.SendDataContext;
import cn.minbb.serialport.impl.SendDataBytesImpl;
import cn.minbb.serialport.impl.SendDataFileImpl;
import cn.minbb.serialport.impl.SendDataStringImpl;
import cn.minbb.serialport.listener.OnDataProgressListener;
import cn.minbb.serialport.listener.OnOpenSerialPortListener;
import cn.minbb.serialport.listener.OnSerialPortDataListener;
import cn.minbb.serialport.thread.SerialPortReadThread;
import cn.minbb.serialport.utils.ByteUtil;
import cn.minbb.serialport.utils.LogUtil;

/**
 * 串口管理器 - 单例模式  防止 反射 克隆 序列化 破坏
 */
public class SerialPortManager extends SerialPort implements Serializable, Cloneable {

    private static final String TAG = SerialPortManager.class.getSimpleName();
    private FileInputStream mFileInputStream;
    private FileOutputStream mFileOutputStream;
    private FileDescriptor mFd;
    private OnOpenSerialPortListener mOnOpenSerialPortListener;
    private OnSerialPortDataListener mOnSerialPortDataListener;
    private OnDataProgressListener onDataProgressListener;

    private HandlerThread mSendingHandlerThread;
    private Handler mSendingHandler;
    private SerialPortReadThread mSerialPortReadThread;

    private SerialPortStatus serialPortStatus = SerialPortStatus.CLOSE;

    // 报文长度
    private static int LENGTH = 64;
    // 超时时间
    private static int TIMEOUT = 5000;

    // 默认是第一次创建
    private static volatile boolean isCreate = false;
    // 单例对象
    private static SerialPortManager serialPortManager = null;

    private SerialPortManager() {
        if (isCreate) {
            throw new RuntimeException("已然被实例化一次，不能在实例化");
        }
        isCreate = true;
    }

    /**
     * 线程安全 双重检查加锁
     */
    public static SerialPortManager getInstance() {
        if (null == serialPortManager) {
            synchronized (SerialPortManager.class) {
                if (null == serialPortManager) {
                    serialPortManager = new SerialPortManager();
                }
            }
        }
        return serialPortManager;
    }

    /**
     * 获取当前串口状态
     */
    public SerialPortStatus getSerialPortStatus() {
        return serialPortStatus;
    }

    /**
     * 设置接收数据超时时间
     *
     * @param timeout 毫秒
     */
    public static void setTimeout(int timeout) {
        SerialPortManager.TIMEOUT = timeout;
    }

    /**
     * 打开串口
     * <p>
     * 串口设备
     * 波特率
     *
     * @return 打开是否成功
     */
    public boolean openDefaultSerialPort() {
        return openCustomSerialPort(new File("/dev/ttyS1"), 115200);
    }

    public boolean openCustomSerialPort(File device, int baudRate) {
        Log.i(TAG, "openSerialPort: " + String.format("打开串口 %s 波特率 %s", device.getPath(), baudRate));
        // 校验串口权限
        if (!device.canRead() || !device.canWrite()) {
            boolean chmod777 = chmod777(device);
            if (!chmod777) {
                Log.i(TAG, "openSerialPort: 没有读写权限");
                if (null != mOnOpenSerialPortListener) {
                    mOnOpenSerialPortListener.onFail(device, OnOpenSerialPortListener.Status.NO_READ_WRITE_PERMISSION);
                }
                return false;
            }
        }

        try {
            mFd = open(device.getAbsolutePath(), baudRate, 0);
            mFileInputStream = new FileInputStream(mFd);
            mFileOutputStream = new FileOutputStream(mFd);
            Log.i(TAG, "openSerialPort: 串口已经打开 " + mFd);
            if (null != mOnOpenSerialPortListener) {
                mOnOpenSerialPortListener.onSuccess(device);
            }
            // 开启发送消息的线程
            startSendThread();
            // 开启接收消息的线程
            startReadThread();
            serialPortStatus = SerialPortStatus.OPEN;
            return true;
        } catch (Exception e) {
            e.printStackTrace();
            if (null != mOnOpenSerialPortListener) {
                mOnOpenSerialPortListener.onFail(device, OnOpenSerialPortListener.Status.OPEN_FAIL);
            }
        }
        return false;
    }

    /**
     * 关闭串口
     */
    public void closeSerialPort() {
        if (null != mFd) {
            close();
            mFd = null;
        }
        // 停止发送消息的线程
        stopSendThread();
        // 停止接收消息的线程
        stopReadThread();

        if (null != mFileInputStream) {
            try {
                mFileInputStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
            mFileInputStream = null;
        }

        if (null != mFileOutputStream) {
            try {
                mFileOutputStream.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
            mFileOutputStream = null;
        }

        mOnOpenSerialPortListener = null;
        mOnSerialPortDataListener = null;
        serialPortStatus = SerialPortStatus.CLOSE;
    }

    /**
     * 添加打开串口监听
     *
     * @param listener listener
     * @return SerialPortManager
     */
    public SerialPortManager setOnOpenSerialPortListener(OnOpenSerialPortListener listener) {
        mOnOpenSerialPortListener = listener;
        return this;
    }

    /**
     * 添加数据通信监听
     *
     * @param listener listener
     * @return SerialPortManager
     */
    public SerialPortManager setOnSerialPortDataListener(OnSerialPortDataListener listener) {
        mOnSerialPortDataListener = listener;
        return this;
    }

    /**
     * 开启发送消息的线程
     */
    private void startSendThread() {
        // 开启发送消息的线程
        mSendingHandlerThread = new HandlerThread("mSendingHandlerThread");
        mSendingHandlerThread.start();
        // Handler
        mSendingHandler = new Handler(mSendingHandlerThread.getLooper()) {
            @Override
            public void handleMessage(Message msg) {
                try {
//                    byte[] sendBytes = new byte[64];
                    Data data = (Data) msg.obj;
                    switch (data.getDataType()) {
                        case BYTES:
                            SendDataBytesImpl sendDataBytesImpl = new SendDataBytesImpl();
                            SendDataContext<byte[]> contextBytes = new SendDataContext<>(mFileInputStream, mFileOutputStream, onDataProgressListener, sendDataBytesImpl);
                            contextBytes.doSend((byte[]) data.getData());
                            break;
                        case STRING:
                            SendDataStringImpl sendDataStringImpl = new SendDataStringImpl();
                            SendDataContext<String> contextString = new SendDataContext<>(mFileInputStream, mFileOutputStream, onDataProgressListener, sendDataStringImpl);
                            contextString.doSend((String) data.getData());
                            break;
                        case FILE:
                            SendDataFileImpl sendDataFileImpl = new SendDataFileImpl();
                            SendDataContext<File> contextFile = new SendDataContext<>(mFileInputStream, mFileOutputStream, onDataProgressListener, sendDataFileImpl);
                            contextFile.doSend((File) data.getData());
                            break;
                    }
//                    if (null != mFileOutputStream && null != sendBytes && 0 < sendBytes.length) {
//                        try {
//                            mFileOutputStream.write(sendBytes);
//                            if (null != mOnSerialPortDataListener) {
//                                mOnSerialPortDataListener.onDataSent(sendBytes);
//                            }
//                        } catch (IOException e) {
//                            e.printStackTrace();
//                        }
//                    }
                } catch (IOException e) {
                    throw new RuntimeException(e);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        };
    }

    /**
     * 停止发送消息线程
     */
    private void stopSendThread() {
        mSendingHandler = null;
        if (null != mSendingHandlerThread) {
            mSendingHandlerThread.interrupt();
            mSendingHandlerThread.quit();
            mSendingHandlerThread = null;
        }
    }

    /**
     * 开启接收消息的线程
     */
    private static byte[] dataBytes = new byte[0];

    private void startReadThread() {
        mSerialPortReadThread = new SerialPortReadThread(mFileInputStream) {
            @Override
            public void onDataReceived(final byte[] bytes) {
                // 收到 EOF(-1) 或者超时则认为数据包接收完毕
                // 串口空闲状态下收到数据，启动超时计时器，并修改串口状态为数据输入状态
                if (SerialPortStatus.OPEN == serialPortStatus) {
                    serialPortStatus = SerialPortStatus.DATA_IN;
                    new Timer().schedule(new TimerTask() {
                        @Override
                        public void run() {
                            // 超时执行
                            if (SerialPortStatus.DATA_IN == serialPortStatus) {
                                // 解包收到的数据
                                LogUtil.setAppendFile("接收超时");
                                dataNotify();
                            }
                        }
                    }, TIMEOUT);
                }
                // 遍历数据块，检测到报文结束符 EOF(-1) 则终止报文接收
                int eofIndex = -1;
                int blockSize = bytes.length;
                byte[] byteBlock = new byte[blockSize];
                for (int i = 0; i < blockSize; i++) {
                    if (bytes[i] == -1) {
                        eofIndex = i;
                        break;
                    } else {
                        byteBlock[i] = bytes[i];
                    }
                }
                if (eofIndex != -1) {
                    // 收到包尾，拼接包尾数据后通知
                    byte[] tailBytes = new byte[eofIndex];
                    System.arraycopy(byteBlock, 0, tailBytes, 0, eofIndex);
                    dataBytesReceived(tailBytes);
                    dataNotify();
                } else {
                    dataBytesReceived(byteBlock);
                }
            }
        };
        mSerialPortReadThread.start();
    }

    /**
     * 停止接收消息的线程
     */
    private void stopReadThread() {
        if (null != mSerialPortReadThread) {
            mSerialPortReadThread.release();
        }
    }

    /**
     * 发送字符数组数据
     *
     * @param sendBytes 字符数组
     */
    public void sendBytes(byte[] sendBytes, OnDataProgressListener onDataProgressListener) {
        this.onDataProgressListener = onDataProgressListener;
        sendData(new Data(sendBytes, DataType.BYTES));
    }

    /**
     * 发送字符串数据（UTF-8编码）
     *
     * @param string 字符串
     */
    public void sendString(String string, OnDataProgressListener onDataProgressListener) {
        Data data = new Data(string, DataType.STRING);
        this.onDataProgressListener = onDataProgressListener;
        sendData(data);
    }

    /**
     * 发送文件数据
     *
     * @param file 文件
     */
    public void sendFile(File file, OnDataProgressListener onDataProgressListener) throws Exception {
        if (!file.exists()) throw new FileNotFoundException("文件不存在");
        if (file.isDirectory()) throw new RuntimeException("文件夹不可以发送");
        this.onDataProgressListener = onDataProgressListener;
        sendData(new Data(file, DataType.FILE));
    }

    /**
     * 发送数据
     *
     * @param data 数据
     * @return 是否发送成功
     */
    private boolean sendData(Data data) {
        if (null != mFd && null != mFileInputStream && null != mFileOutputStream) {
            if (null != mSendingHandler) {
                Message message = Message.obtain();
                message.obj = data;
                return mSendingHandler.sendMessage(message);
            }
            LogUtil.setAppendFile("缺失发送处理器");
            throw new RuntimeException("缺失发送处理器");
        }
        LogUtil.setAppendFile("数据流为空");
        throw new RuntimeException("数据流为空");
    }

    /**
     * 数据回调方法，当串口接收数据超时或者收到数据结束符时候调用，用于通知应用程序收到的数据
     */
    private void dataNotify() {
        // 串口状态设置为空闲
        serialPortStatus = SerialPortStatus.OPEN;
        if (null != mOnSerialPortDataListener) {
            mOnSerialPortDataListener.onDataReceived(dataBytes, new String(dataBytes));
            LogUtil.setAppendFile("Notify:" + new String(dataBytes) + Arrays.toString(dataBytes));
            dataBytes = new byte[0];
        } else {
            LogUtil.setAppendFile("串口数据监听器未定义");
            throw new RuntimeException("串口数据监听器未定义");
        }
    }

    /**
     * 收到串口数据调用，同步锁
     *
     * @param bytes 收到的 Byte 数组
     */
    private synchronized void dataBytesReceived(byte[] bytes) {
        dataBytes = ByteUtil.addBytes(dataBytes, bytes);
    }

    @Override
    protected Object clone() throws CloneNotSupportedException {
        return serialPortManager;
    }

    /**
     * 防止序列化破坏
     */
    private Object readResolve() {
        return serialPortManager;
    }
}
